@@ -25,6 +25,7 @@ from dracclient.resources import idrac_card
 from dracclient.resources import inventory
 from dracclient.resources import job
 from dracclient.resources import lifecycle_controller
+from dracclient.resources import nic
 from dracclient.resources import raid
 from dracclient.resources import system
 from dracclient.resources import uris
@@ -40,6 +41,7 @@ class DRACClient(object):
     """Client for managing DRAC nodes"""
 
     BIOS_DEVICE_FQDD = 'BIOS.Setup.1-1'
+    IDRAC_FQDD = 'iDRAC.Embedded.1'
 
     def __init__(
             self, host, username, password, port=443, path='/wsman',
@@ -77,6 +79,7 @@ class DRACClient(object):
         self._raid_mgmt = raid.RAIDManagement(self.client)
         self._system_cfg = system.SystemConfiguration(self.client)
         self._inventory_mgmt = inventory.InventoryManagement(self.client)
+        self._nic_cfg = nic.NICConfiguration(self.client)
 
     def get_power_state(self):
         """Returns the current power state of the node
@@ -170,9 +173,18 @@ class DRACClient(object):
         :param settings: a dictionary containing the proposed values, with
                          each key being the name of attribute and the value
                          being the proposed value.
-        :returns: a dictionary containing the commit_needed key with a boolean
-                  value indicating whether a config job must be created for the
-                  values to be applied.
+        :returns: a dictionary containing:
+                 - The commit_required key with a boolean value indicating
+                   whether a config job must be created for the values to be
+                   applied.  This key actually has a value that indicates if
+                   a reboot is required.  This key has been deprecated and
+                   will be removed in a future release.
+                 - The is_commit_required key with a boolean value indicating
+                   whether a config job must be created for the values to be
+                   applied.
+                 - The is_reboot_required key with a RebootRequired enumerated
+                   value indicating whether the server must be rebooted for the
+                   values to be applied.  Possible values are true and false.
         :raises: WSManRequestFailure on request failures
         :raises: WSManInvalidResponse when receiving invalid response
         :raises: DRACOperationFailed on error reported back by the DRAC
@@ -182,19 +194,101 @@ class DRACClient(object):
         """
         return self._bios_cfg.set_bios_settings(settings)
 
-    def list_idrac_settings(self):
+    def list_idrac_settings(self, by_name=False, fqdd_filter=IDRAC_FQDD):
         """List the iDRAC configuration settings
 
-        :returns: a dictionary with the iDRAC settings using InstanceID as the
-                  key. The attributes are either iDRACCArdEnumerableAttribute,
-                  iDRACCardStringAttribute or iDRACCardIntegerAttribute
-                  objects.
+        :param by_name: Controls whether returned dictionary uses iDRAC card
+                        attribute name as key. If set to False, instance_id
+                        will be used.  If set to True the keys will be of the
+                        form "group_id#name".
+        :param fqdd_filter: An FQDD used to filter the instances.  Note that
+                            this is only used when by_name is True.
+        :returns: a dictionary with the iDRAC settings using instance_id as the
+                  key except when by_name is True. The attributes are either
+                  iDRACCardEnumerableAttribute, iDRACCardStringAttribute or
+                  iDRACCardIntegerAttribute objects.
         :raises: WSManRequestFailure on request failures
         :raises: WSManInvalidResponse when receiving invalid response
         :raises: DRACOperationFailed on error reported back by the DRAC
                  interface
         """
-        return self._idrac_cfg.list_idrac_settings()
+        return self._idrac_cfg.list_idrac_settings(by_name=by_name,
+                                                   fqdd_filter=fqdd_filter)
+
+    def set_idrac_settings(self, settings, idrac_fqdd=IDRAC_FQDD):
+        """Sets the iDRAC configuration settings
+
+        To be more precise, it sets the pending_value parameter for each of the
+        attributes passed in. For the values to be applied, a config job may
+        need to be created and the node may need to be rebooted.
+
+        :param settings: a dictionary containing the proposed values, with
+                         each key being the name of attribute qualified with
+                         the group ID in the form "group_id#name" and the value
+                         being the proposed value.
+        :param idrac_fqdd: the FQDD of the iDRAC.
+        :returns: a dictionary containing:
+                 - The is_commit_required key with a boolean value indicating
+                   whether a config job must be created for the values to be
+                   applied.
+                 - The is_reboot_required key with a RebootRequired enumerated
+                   value indicating whether the server must be rebooted for the
+                   values to be applied.  Possible values are true and false.
+        :raises: WSManRequestFailure on request failures
+        :raises: WSManInvalidResponse when receiving invalid response
+        :raises: DRACOperationFailed on error reported back by the DRAC
+                 interface
+        :raises: DRACUnexpectedReturnValue on return value mismatch
+        :raises: InvalidParameterValue on invalid attribute
+        """
+        return self._idrac_cfg.set_idrac_settings(settings, idrac_fqdd)
+
+    def commit_pending_idrac_changes(
+            self,
+            idrac_fqdd=IDRAC_FQDD,
+            reboot=False,
+            start_time='TIME_NOW'):
+        """Create a config job for applying all pending iDRAC changes.
+
+        :param idrac_fqdd: the FQDD of the iDRAC.
+        :param reboot: indication of whether to also create a reboot job
+        :param start_time: start time for job execution in format
+                           yyyymmddhhmmss, the string 'TIME_NOW' which
+                           means execute immediately or None which means
+                           the job will not execute until
+                           schedule_job_execution is called
+        :returns: id of the created configuration job
+        :raises: WSManRequestFailure on request failures
+        :raises: WSManInvalidResponse when receiving invalid response
+        :raises: DRACOperationFailed on error reported back by the iDRAC
+                 interface
+        :raises: DRACUnexpectedReturnValue on return value mismatch
+        """
+        return self._job_mgmt.create_config_job(
+            resource_uri=uris.DCIM_iDRACCardService,
+            cim_creation_class_name='DCIM_iDRACCardService',
+            cim_name='DCIM:iDRACCardService',
+            target=idrac_fqdd,
+            reboot=reboot,
+            start_time=start_time)
+
+    def abandon_pending_idrac_changes(self, idrac_fqdd=IDRAC_FQDD):
+        """Abandon all pending changes to an iDRAC
+
+        Once a config job has been submitted, it can no longer be abandoned.
+
+        :param idrac_fqdd: the FQDD of the iDRAC.
+        :raises: WSManRequestFailure on request failures
+        :raises: WSManInvalidResponse when receiving invalid response
+        :raises: DRACOperationFailed on error reported back by the iDRAC
+                 interface
+        :raises: DRACUnexpectedReturnValue on return value mismatch
+        """
+        self._job_mgmt.delete_pending_config(
+            resource_uri=uris.DCIM_iDRACCardService,
+            cim_creation_class_name='DCIM_iDRACCardService',
+            cim_name='DCIM:iDRACCardService',
+            target=idrac_fqdd)
 
     def list_lifecycle_settings(self):
         """List the Lifecycle Controller configuration settings
@@ -247,12 +341,37 @@ class DRACClient(object):
         """
         return self._job_mgmt.get_job(job_id)
 
-    def create_config_job(self, resource_uri, cim_creation_class_name,
-                          cim_name, target,
+    def delete_jobs(self, job_ids=['JID_CLEARALL']):
+        """Deletes the given jobs, or all jobs if none specified
+
+        :param job_ids: a list of job ids to delete. Clearing all the
+                jobs may be accomplished using the keyword JID_CLEARALL
+                as the job_id, or JID_CLEARALL_FORCE if a job is in
+                Scheduled state and there is another job for the same
+                component in Completed or Failed state,
+                (http://en.community.dell.com/techcenter/extras/m/white_papers/20444501/download)
+                Deletion of each job id will be attempted, even if there
+                are errors in deleting any in the list.
+        :raises: WSManRequestFailure on request failures
+        :raises: WSManInvalidResponse when receiving invalid response
+        :raises: DRACOperationFailed on error reported back by the iDRAC
+                 interface.  There will be one message for each job_id
+                 that had a failure in the exception.
+        :raises: DRACUnexpectedReturnValue on non-success
+        """
+
+        return self._job_mgmt.delete_jobs(job_ids)
+
+    def create_config_job(self,
+                          resource_uri,
+                          cim_creation_class_name,
+                          cim_name,
+                          target,
                           cim_system_creation_class_name='DCIM_ComputerSystem',
                           cim_system_name='DCIM:ComputerSystem',
-                          reboot=False):
-        """Creates a config job
+                          reboot=False,
+                          start_time='TIME_NOW'):
+        """Creates a configuration job.
 
         In CIM (Common Information Model), weak association is used to name an
         instance of one class in the context of an instance of another class.
@@ -268,18 +387,90 @@ class DRACClient(object):
         :param cim_system_creation_class_name: creation class name of the
                                                scoping system
         :param cim_system_name: name of the scoping system
-        :param reboot: indicates whether a RebootJob should also be
-                       created or not
+        :param reboot: indicates whether or not a RebootJob should also be
+                       created
+        :param start_time: start time for job execution in format
+                           yyyymmddhhmmss, the string 'TIME_NOW' which
+                           means execute immediately or None which means
+                           the job will not execute until
+                           schedule_job_execution is called
         :returns: id of the created job
         :raises: WSManRequestFailure on request failures
         :raises: WSManInvalidResponse when receiving invalid response
-        :raises: DRACOperationFailed on error reported back by the DRAC
+        :raises: DRACOperationFailed on error reported back by the iDRAC
+                 interface
+        :raises: DRACUnexpectedReturnValue on return value mismatch
+        """
+
+        return self._job_mgmt.create_config_job(
+            resource_uri=resource_uri,
+            cim_creation_class_name=cim_creation_class_name,
+            cim_name=cim_name,
+            target=target,
+            cim_system_creation_class_name=cim_system_creation_class_name,
+            cim_system_name=cim_system_name,
+            reboot=reboot,
+            start_time=start_time)
+
+    def create_nic_config_job(
+            self,
+            nic_id,
+            reboot=False,
+            start_time='TIME_NOW'):
+        """Creates config job for applying pending changes to a NIC.
+
+        :param nic_id: id of the network interface controller (NIC)
+        :param reboot: indication of whether to also create a reboot job
+        :param start_time: start time for job execution in format
+                           yyyymmddhhmmss; the string 'TIME_NOW' means
+                           immediately and None means unspecified
+        :returns: id of the created configuration job
+        :raises: WSManRequestFailure on request failures
+        :raises: WSManInvalidResponse when receiving invalid response
+        :raises: DRACOperationFailed on error reported back by the iDRAC
                  interface
         :raises: DRACUnexpectedReturnValue on return value mismatch
         """
         return self._job_mgmt.create_config_job(
-            resource_uri, cim_creation_class_name, cim_name, target,
-            cim_system_creation_class_name, cim_system_name, reboot)
+            resource_uri=uris.DCIM_NICService,
+            cim_creation_class_name='DCIM_NICService',
+            cim_name='DCIM:NICService',
+            target=nic_id,
+            reboot=reboot,
+            start_time=start_time)
+
+    def create_reboot_job(
+            self,
+            reboot_type=constants.RebootJobType.reboot_forced_shutdown):
+        """Creates a reboot job.
+
+        :param reboot_type: type of reboot
+        :returns id of the created job
+        :raises: InvalidParameterValue on invalid reboot type
+        :raises: WSManRequestFailure on request failures
+        :raises: WSManInvalidResponse when receiving invalid response
+        :raises: DRACOperationFailed on error reported back by the iDRAC
+                 interface
+        :raises: DRACUnexpectedReturnValue on return value mismatch
+        """
+        return self._job_mgmt.create_reboot_job(reboot_type)
+
+    def schedule_job_execution(self, job_ids, start_time='TIME_NOW'):
+        """Schedules jobs for execution in a specified order.
+
+        :param job_ids: list of job identifiers
+        :param start_time: start time for job execution in format
+                           yyyymmddhhmmss or the string 'TIME_NOW' which
+                           means execute immediately.  None is not a
+                           valid option for this request.
+        :raises: WSManRequestFailure on request failures
+        :raises: WSManInvalidResponse when receiving invalid response
+        :raises: DRACOperationFailed on error reported back by the iDRAC
+                 interface, including start_time being in the past or
+                 badly formatted start_time
+        :raises: DRACUnexpectedReturnValue on return value mismatch
+        """
+        return self._job_mgmt.schedule_job_execution(job_ids, start_time)
 
     def delete_pending_config(
             self, resource_uri, cim_creation_class_name, cim_name, target,
@@ -313,23 +504,34 @@ class DRACClient(object):
             resource_uri, cim_creation_class_name, cim_name, target,
             cim_system_creation_class_name, cim_system_name)
 
-    def commit_pending_bios_changes(self, reboot=False):
+    def commit_pending_bios_changes(
+            self,
+            reboot=False,
+            start_time='TIME_NOW'):
         """Applies all pending changes on the BIOS by creating a config job
 
         :param reboot: indicates whether a RebootJob should also be
                        created or not
+        :param start_time: start time for job execution in format
+                           yyyymmddhhmmss, the string 'TIME_NOW' which
+                           means execute immediately or None which means
+                           the job will not execute until
+                           schedule_job_execution is called
         :returns: id of the created job
         :raises: WSManRequestFailure on request failures
         :raises: WSManInvalidResponse when receiving invalid response
         :raises: DRACOperationFailed on error reported back by the DRAC
-                 interface
+                 interface, including start_time being in the past or
+                 badly formatted start_time
         :raises: DRACUnexpectedReturnValue on return value mismatch
         """
         return self._job_mgmt.create_config_job(
             resource_uri=uris.DCIM_BIOSService,
             cim_creation_class_name='DCIM_BIOSService',
-            cim_name='DCIM:BIOSService', target=self.BIOS_DEVICE_FQDD,
-            reboot=reboot)
+            cim_name='DCIM:BIOSService',
+            target=self.BIOS_DEVICE_FQDD,
+            reboot=reboot,
+            start_time=start_time)
 
     def abandon_pending_bios_changes(self):
         """Deletes all pending changes on the BIOS
@@ -404,9 +606,18 @@ class DRACClient(object):
         :param raid_enable: boolean flag, set to True if the disk is to
                become part of the RAID.  The same flag is applied to all
                listed disks
-        :returns: a dictionary containing the commit_required key with a
-                  boolean value indicating whether a config job must be
-                  created for the values to be applied.
+        :returns: a dictionary containing:
+                 - The commit_required key with a boolean value indicating
+                   whether a config job must be created for the values to be
+                   applied.  This key actually has a value that indicates if
+                   a reboot is required.  This key has been deprecated and
+                   will be removed in a future release.
+                 - The is_commit_required key with the value always set to
+                   True indicating that a config job must be created to
+                   complete disk conversion.
+                 - The is_reboot_required key with a RebootRequired enumerated
+                   value indicating whether the server must be rebooted to
+                   complete disk conversion.
         """
         return self._raid_mgmt.convert_physical_disks(
             physical_disks, raid_enable)
@@ -425,9 +636,18 @@ class DRACClient(object):
         :param disk_name: name of the virtual disk (optional)
         :param span_length: number of disks per span (optional)
         :param span_depth: number of spans in virtual disk (optional)
-        :returns: a dictionary containing the commit_needed key with a boolean
-                  value indicating whether a config job must be created for the
-                  values to be applied.
+        :returns: a dictionary containing:
+                 - The commit_required key with a boolean value indicating
+                   whether a config job must be created for the values to be
+                   applied.  This key actually has a value that indicates if
+                   a reboot is required.  This key has been deprecated and
+                   will be removed in a future release.
+                 - The is_commit_required key with the value always set to
+                   True indicating that a config job must be created to
+                   complete virtual disk creation.
+                 - The is_reboot_required key with a RebootRequired enumerated
+                   value indicating whether the server must be rebooted to
+                   complete virtual disk creation.
         :raises: WSManRequestFailure on request failures
         :raises: WSManInvalidResponse when receiving invalid response
         :raises: DRACOperationFailed on error reported back by the DRAC
@@ -446,9 +666,18 @@ class DRACClient(object):
         be applied, a config job must be created and the node must be rebooted.
 
         :param virtual_disk: id of the virtual disk
-        :returns: a dictionary containing the commit_needed key with a boolean
-                  value indicating whether a config job must be created for the
-                  values to be applied.
+        :returns: a dictionary containing:
+                 - The commit_required key with a boolean value indicating
+                   whether a config job must be created for the values to be
+                   applied.  This key actually has a value that indicates if
+                   a reboot is required.  This key has been deprecated and
+                   will be removed in a future release.
+                 - The is_commit_required key with the value always set to
+                   True indicating that a config job must be created to
+                   complete virtual disk deletion.
+                 - The is_reboot_required key with a RebootRequired enumerated
+                   value indicating whether the server must be rebooted to
+                   complete virtual disk deletion.
         :raises: WSManRequestFailure on request failures
         :raises: WSManInvalidResponse when receiving invalid response
         :raises: DRACOperationFailed on error reported back by the DRAC
@@ -457,7 +686,8 @@ class DRACClient(object):
         """
         return self._raid_mgmt.delete_virtual_disk(virtual_disk)
 
-    def commit_pending_raid_changes(self, raid_controller, reboot=False):
+    def commit_pending_raid_changes(self, raid_controller, reboot=False,
+                                    start_time='TIME_NOW'):
         """Applies all pending changes on a RAID controller
 
          ...by creating a config job.
@@ -465,6 +695,11 @@ class DRACClient(object):
         :param raid_controller: id of the RAID controller
         :param reboot: indicates whether a RebootJob should also be
                        created or not
+        :param start_time: start time for job execution in format
+               yyyymmddhhmmss, the string 'TIME_NOW' which
+               means execute immediately or None which means
+               the job will not execute until
+               schedule_job_execution is called
         :returns: id of the created job
         :raises: WSManRequestFailure on request failures
         :raises: WSManInvalidResponse when receiving invalid response
@@ -475,7 +710,10 @@ class DRACClient(object):
         return self._job_mgmt.create_config_job(
             resource_uri=uris.DCIM_RAIDService,
             cim_creation_class_name='DCIM_RAIDService',
-            cim_name='DCIM:RAIDService', target=raid_controller, reboot=reboot)
+            cim_name='DCIM:RAIDService',
+            target=raid_controller,
+            reboot=reboot,
+            start_time=start_time)
 
     def abandon_pending_raid_changes(self, raid_controller):
         """Deletes all pending changes on a RAID controller
@@ -517,9 +755,10 @@ class DRACClient(object):
 
         return self._inventory_mgmt.list_memory()
 
-    def list_nics(self):
+    def list_nics(self, sort=False):
         """Returns a list of NICs
 
+        :param sort: indicates if the list should be sorted or not.
         :returns: a list of NIC objects
         :raises: WSManRequestFailure on request failures
         :raises: WSManInvalidResponse when receiving invalid response
@@ -527,7 +766,53 @@ class DRACClient(object):
                  interface
         """
 
-        return self._inventory_mgmt.list_nics()
+        return self._inventory_mgmt.list_nics(sort=sort)
+
+    def list_nic_settings(self, nic_id):
+        """Return the list of attribute settings of a NIC.
+
+        :param nic_id: id of the network interface controller (NIC)
+        :returns: dictionary containing the NIC settings. The keys are
+                  attribute names. Each value is a
+                  NICEnumerationAttribute, NICIntegerAttribute, or
+                  NICStringAttribute object.
+        :raises: WSManRequestFailure on request failures
+        :raises: WSManInvalidResponse when receiving invalid response
+        :raises: DRACOperationFailed on error reported back by the iDRAC
+                 interface
+        """
+        return self._nic_cfg.list_nic_settings(nic_id)
+
+    def set_nic_settings(self, nic_id, settings):
+        """Modify one or more settings of a NIC.
+
+        If successful, the pending values of the attributes are set. For
+        the new values to be applied, a configuration job must be
+        created and the node must be rebooted.
+
+        :param nic_id: id of the network interface controller (NIC)
+        :param settings: dictionary containing the proposed values, with
+                         each key being the name of an attribute and the
+                         value being the proposed value
+        :returns: dictionary containing a 'commit_required' key with a
+                  boolean value indicating whether a configuration job
+                  must be created for the new settings to be applied
+        :raises: WSManRequestFailure on request failures
+        :raises: WSManInvalidResponse when receiving invalid response
+        :raises: DRACOperationFailed on error reported back by the iDRAC
+                 interface
+        :raises: InvalidParameterValue on invalid NIC attribute
+        """
+        return self._nic_cfg.set_nic_settings(nic_id, settings)
+
+    def get_system(self):
+        """Return a Systen object.
+
+        :returns: a System object
+        :raises: WSManRequestFailure on request failures
+        :raises: WSManInvalidResponse when receiving invalid response
+        """
+        return self._inventory_mgmt.get_system()
 
     def is_idrac_ready(self):
         """Indicates if the iDRAC is ready to accept commands
@@ -565,6 +850,20 @@ class DRACClient(object):
 
         return self.client.wait_until_idrac_is_ready(retries, retry_delay)
 
+    def is_jbod_capable(self, raid_controller_fqdd):
+        """Find out if raid controller supports jbod
+
+        :param raid_controller_fqdd: The raid controller's fqdd
+                        being being checked to see if it is jbod
+                        capable.
+        :raises: DRACRequestFailed if unable to find any disks in the Ready
+                 or non-RAID states
+        :raises: DRACOperationFailed on error reported back by the DRAC
+                 and the exception message does not contain
+                 NOT_SUPPORTED_MSG constant
+        """
+        return self._raid_mgmt.is_jbod_capable(raid_controller_fqdd)
+
 
 class WSManClient(wsman.Client):
     """Wrapper for wsman.Client that can wait until iDRAC is ready
@@ -632,8 +931,14 @@ class WSManClient(wsman.Client):
                                                   max_elems, auto_pull,
                                                   filter_query, filter_dialect)
 
-    def invoke(self, resource_uri, method, selectors=None, properties=None,
-               expected_return_value=None, wait_for_idrac=True):
+    def invoke(self,
+               resource_uri,
+               method,
+               selectors=None,
+               properties=None,
+               expected_return_value=None,
+               wait_for_idrac=True,
+               check_return_value=True):
         """Invokes a remote WS-Man method
 
         :param resource_uri: URI of the resource
@@ -647,6 +952,8 @@ class WSManClient(wsman.Client):
         :param wait_for_idrac: indicates whether or not to wait for the
             iDRAC to be ready to accept commands before issuing the
             command
+        :param check_return_value: indicates if the ReturnValue should be
+            checked and an exception thrown on an unexpected value
         :returns: an lxml.etree.Element object of the response received
         :raises: WSManRequestFailure on request failures
         :raises: WSManInvalidResponse when receiving invalid response
@@ -666,17 +973,21 @@ class WSManClient(wsman.Client):
         resp = super(WSManClient, self).invoke(resource_uri, method, selectors,
                                                properties)
 
-        return_value = utils.find_xml(resp, 'ReturnValue', resource_uri).text
-        if return_value == utils.RET_ERROR:
-            message_elems = utils.find_xml(resp, 'Message', resource_uri, True)
-            messages = [message_elem.text for message_elem in message_elems]
-            raise exceptions.DRACOperationFailed(drac_messages=messages)
+        if check_return_value:
+            return_value = utils.find_xml(resp, 'ReturnValue',
+                                          resource_uri).text
+            if return_value == utils.RET_ERROR:
+                message_elems = utils.find_xml(resp, 'Message',
+                                               resource_uri, True)
+                messages = [message_elem.text for message_elem in
+                            message_elems]
+                raise exceptions.DRACOperationFailed(drac_messages=messages)
 
-        if (expected_return_value is not None and
-                return_value != expected_return_value):
-            raise exceptions.DRACUnexpectedReturnValue(
-                expected_return_value=expected_return_value,
-                actual_return_value=return_value)
+            if (expected_return_value is not None and
+                    return_value != expected_return_value):
+                raise exceptions.DRACUnexpectedReturnValue(
+                    expected_return_value=expected_return_value,
+                    actual_return_value=return_value)
 
         return resp
 
