@@ -66,14 +66,17 @@ PHYSICAL_DISK_BUS_PROTOCOL = {
     '3': 'fibre',
     '4': 'usb',
     '5': 'sata',
-    '6': 'sas'
+    '6': 'sas',
+    '7': 'pcie',
+    '8': 'nvme'
 }
 
 PhysicalDiskTuple = collections.namedtuple(
     'PhysicalDisk',
     ['id', 'description', 'controller', 'manufacturer', 'model', 'media_type',
      'interface_type', 'size_mb', 'free_size_mb', 'serial_number',
-     'firmware_version', 'status', 'raid_status'])
+     'firmware_version', 'status', 'raid_status', 'sas_address',
+     'device_protocol'])
 
 
 class PhysicalDisk(PhysicalDiskTuple):
@@ -107,7 +110,7 @@ class PhysicalDisk(PhysicalDiskTuple):
 
 RAIDController = collections.namedtuple(
     'RAIDController', ['id', 'description', 'manufacturer', 'model',
-                       'primary_status', 'firmware_version'])
+                       'primary_status', 'firmware_version', 'bus'])
 
 VirtualDiskTuple = collections.namedtuple(
     'VirtualDisk',
@@ -148,6 +151,8 @@ class VirtualDisk(VirtualDiskTuple):
 
 class RAIDManagement(object):
 
+    NOT_SUPPORTED_MSG = " operation is not supported on th"
+
     def __init__(self, client):
         """Creates RAIDManagement object
 
@@ -187,7 +192,8 @@ class RAIDManagement(object):
                 self._get_raid_controller_attr(drac_controller,
                                                'PrimaryStatus')],
             firmware_version=self._get_raid_controller_attr(
-                drac_controller, 'ControllerFirmwareVersion'))
+                drac_controller, 'ControllerFirmwareVersion'),
+            bus=self._get_raid_controller_attr(drac_controller, 'Bus'))
 
     def _get_raid_controller_attr(self, drac_controller, attr_name):
         return utils.get_wsman_resource_attr(
@@ -231,7 +237,7 @@ class RAIDManagement(object):
                                                     nullable=True),
             controller=fqdd.split(':')[-1],
             raid_level=REVERSE_RAID_LEVELS[drac_raid_level],
-            size_mb=int(size_b) / 2 ** 20,
+            size_mb=int(size_b) // 2 ** 20,
             status=constants.PRIMARY_STATUS[drac_status],
             raid_status=DISK_RAID_STATUS[drac_raid_status],
             span_depth=int(self._get_virtual_disk_attr(drac_disk,
@@ -267,44 +273,74 @@ class RAIDManagement(object):
         drac_physical_disks = utils.find_xml(doc, 'DCIM_PhysicalDiskView',
                                              uris.DCIM_PhysicalDiskView,
                                              find_all=True)
+        physical_disks = [self._parse_drac_physical_disk(disk)
+                          for disk in drac_physical_disks]
 
-        return [self._parse_drac_physical_disk(disk)
-                for disk in drac_physical_disks]
+        drac_pcie_disks = utils.find_xml(doc, 'DCIM_PCIeSSDView',
+                                         uris.DCIM_PCIeSSDView,
+                                         find_all=True)
+        pcie_disks = [self._parse_drac_physical_disk(disk,
+                      uris.DCIM_PCIeSSDView) for disk in drac_pcie_disks]
+
+        return physical_disks + pcie_disks
+
+    def _parse_drac_physical_disk(self,
+                                  drac_disk,
+                                  uri=uris.DCIM_PhysicalDiskView):
+        fqdd = self._get_physical_disk_attr(drac_disk, 'FQDD', uri)
+        size_b = self._get_physical_disk_attr(drac_disk, 'SizeInBytes', uri)
 
-    def _parse_drac_physical_disk(self, drac_disk):
-        fqdd = self._get_physical_disk_attr(drac_disk, 'FQDD')
-        size_b = self._get_physical_disk_attr(drac_disk, 'SizeInBytes')
         free_size_b = self._get_physical_disk_attr(drac_disk,
-                                                   'FreeSizeInBytes')
-        drac_status = self._get_physical_disk_attr(drac_disk, 'PrimaryStatus')
+                                                   'FreeSizeInBytes', uri)
+        if free_size_b is not None:
+            free_size_mb = int(free_size_b) // 2 ** 20
+        else:
+            free_size_mb = None
+
+        drac_status = self._get_physical_disk_attr(drac_disk, 'PrimaryStatus',
+                                                   uri)
         drac_raid_status = self._get_physical_disk_attr(drac_disk,
-                                                        'RaidStatus')
-        drac_media_type = self._get_physical_disk_attr(drac_disk, 'MediaType')
+                                                        'RaidStatus', uri)
+        if drac_raid_status is not None:
+            raid_status = DISK_RAID_STATUS[drac_raid_status]
+        else:
+            raid_status = None
+        drac_media_type = self._get_physical_disk_attr(drac_disk, 'MediaType',
+                                                       uri)
         drac_bus_protocol = self._get_physical_disk_attr(drac_disk,
-                                                         'BusProtocol')
+                                                         'BusProtocol', uri)
 
         return PhysicalDisk(
             id=fqdd,
             description=self._get_physical_disk_attr(drac_disk,
-                                                     'DeviceDescription'),
+                                                     'DeviceDescription',
+                                                     uri),
             controller=fqdd.split(':')[-1],
             manufacturer=self._get_physical_disk_attr(drac_disk,
-                                                      'Manufacturer'),
-            model=self._get_physical_disk_attr(drac_disk, 'Model'),
+                                                      'Manufacturer', uri),
+            model=self._get_physical_disk_attr(drac_disk, 'Model', uri),
             media_type=PHYSICAL_DISK_MEDIA_TYPE[drac_media_type],
             interface_type=PHYSICAL_DISK_BUS_PROTOCOL[drac_bus_protocol],
-            size_mb=int(size_b) / 2 ** 20,
-            free_size_mb=int(free_size_b) / 2 ** 20,
+            size_mb=int(size_b) // 2 ** 20,
+            free_size_mb=free_size_mb,
             serial_number=self._get_physical_disk_attr(drac_disk,
-                                                       'SerialNumber'),
+                                                       'SerialNumber', uri),
             firmware_version=self._get_physical_disk_attr(drac_disk,
-                                                          'Revision'),
+                                                          'Revision', uri),
             status=constants.PRIMARY_STATUS[drac_status],
-            raid_status=DISK_RAID_STATUS[drac_raid_status])
+            raid_status=raid_status,
+            sas_address=self._get_physical_disk_attr(drac_disk, 'SASAddress',
+                                                     uri, allow_missing=True),
+            device_protocol=self._get_physical_disk_attr(drac_disk,
+                                                         'DeviceProtocol',
+                                                         uri,
+                                                         allow_missing=True))
 
-    def _get_physical_disk_attr(self, drac_disk, attr_name):
+    def _get_physical_disk_attr(self, drac_disk, attr_name, uri,
+                                allow_missing=False):
         return utils.get_wsman_resource_attr(
-            drac_disk, uris.DCIM_PhysicalDiskView, attr_name, nullable=True)
+            drac_disk, uri, attr_name, nullable=True,
+            allow_missing=allow_missing)
 
     def convert_physical_disks(self, physical_disks, raid_enable):
         """Converts a list of physical disks into or out of RAID mode.
@@ -316,9 +352,18 @@ class RAIDManagement(object):
         :param raid_enable: boolean flag, set to True if the disk is to
                become part of the RAID.  The same flag is applied to all
                listed disks
-        :returns: a dictionary containing the commit_needed key with a boolean
-                  value indicating whether a config job must be created for the
-                  values to be applied.
+        :returns: a dictionary containing:
+                 - The commit_required key with a boolean value indicating
+                   whether a config job must be created for the values to be
+                   applied.  This key actually has a value that indicates if
+                   a reboot is required.  This key has been deprecated and
+                   will be removed in a future release.
+                 - The is_commit_required key with the value always set to
+                   True indicating that a config job must be created to
+                   complete disk conversion.
+                 - The is_reboot_required key with a RebootRequired enumerated
+                   value indicating whether the server must be rebooted to
+                   complete disk conversion.
         """
         invocation = 'ConvertToRAID' if raid_enable else 'ConvertToNonRAID'
 
@@ -333,8 +378,9 @@ class RAIDManagement(object):
                                  selectors, properties,
                                  expected_return_value=utils.RET_SUCCESS)
 
-        return {'commit_required':
-                utils.is_reboot_required(doc, uris.DCIM_RAIDService)}
+        return utils.build_return_dict(doc, uris.DCIM_RAIDService,
+                                       include_commit_required=True,
+                                       is_commit_required_value=True)
 
     def create_virtual_disk(self, raid_controller, physical_disks, raid_level,
                             size_mb, disk_name=None, span_length=None,
@@ -351,9 +397,18 @@ class RAIDManagement(object):
         :param disk_name: name of the virtual disk (optional)
         :param span_length: number of disks per span (optional)
         :param span_depth: number of spans in virtual disk (optional)
-        :returns: a dictionary containing the commit_needed key with a boolean
-                  value indicating whether a config job must be created for the
-                  values to be applied.
+        :returns: a dictionary containing:
+                 - The commit_required key with a boolean value indicating
+                   whether a config job must be created for the values to be
+                   applied.  This key actually has a value that indicates if
+                   a reboot is required.  This key has been deprecated and
+                   will be removed in a future release.
+                 - The is_commit_required key with the value always set to
+                   True indicating that a config job must be created to
+                   complete virtual disk creation.
+                 - The is_reboot_required key with a RebootRequired enumerated
+                   value indicating whether the server must be rebooted to
+                   complete virtual disk creation.
         :raises: WSManRequestFailure on request failures
         :raises: WSManInvalidResponse when receiving invalid response
         :raises: DRACOperationFailed on error reported back by the DRAC
@@ -375,10 +430,7 @@ class RAIDManagement(object):
             error_msgs.append("'physical_disks' is not supplied")
 
         # size validation
-        if not size_mb:
-            error_msgs.append("'size_mb' is not supplied")
-        else:
-            utils.validate_integer_value(size_mb, 'size_mb', error_msgs)
+        utils.validate_integer_value(size_mb, 'size_mb', error_msgs)
 
         virtual_disk_prop_names.append('Size')
         virtual_disk_prop_values.append(str(size_mb))
@@ -424,8 +476,9 @@ class RAIDManagement(object):
                                  selectors, properties,
                                  expected_return_value=utils.RET_SUCCESS)
 
-        return {'commit_required': utils.is_reboot_required(
-            doc, uris.DCIM_RAIDService)}
+        return utils.build_return_dict(doc, uris.DCIM_RAIDService,
+                                       include_commit_required=True,
+                                       is_commit_required_value=True)
 
     def delete_virtual_disk(self, virtual_disk):
         """Deletes a virtual disk
@@ -434,9 +487,18 @@ class RAIDManagement(object):
         be applied, a config job must be created and the node must be rebooted.
 
         :param virtual_disk: id of the virtual disk
-        :returns: a dictionary containing the commit_needed key with a boolean
-                  value indicating whether a config job must be created for the
-                  values to be applied.
+        :returns: a dictionary containing:
+                 - The commit_required key with a boolean value indicating
+                   whether a config job must be created for the values to be
+                   applied.  This key actually has a value that indicates if
+                   a reboot is required.  This key has been deprecated and
+                   will be removed in a future release.
+                 - The is_commit_required key with the value always set to
+                   True indicating that a config job must be created to
+                   complete virtual disk deletion.
+                 - The is_reboot_required key with a RebootRequired enumerated
+                   value indicating whether the server must be rebooted to
+                   complete virtual disk deletion.
         :raises: WSManRequestFailure on request failures
         :raises: WSManInvalidResponse when receiving invalid response
         :raises: DRACOperationFailed on error reported back by the DRAC
@@ -454,5 +516,64 @@ class RAIDManagement(object):
                                  selectors, properties,
                                  expected_return_value=utils.RET_SUCCESS)
 
-        return {'commit_required': utils.is_reboot_required(
-            doc, uris.DCIM_RAIDService)}
+        return utils.build_return_dict(doc, uris.DCIM_RAIDService,
+                                       include_commit_required=True,
+                                       is_commit_required_value=True)
+
+    def is_jbod_capable(self, raid_controller_fqdd):
+        """Find out if raid controller supports jbod
+
+        :param raid_controller_fqdd: The raid controller's fqdd
+                        being being checked to see if it is jbod
+                        capable.
+        :raises: DRACRequestFailed if unable to find any disks in the Ready
+                 or non-RAID states
+        :raises: DRACOperationFailed on error reported back by the DRAC
+                 and the exception message does not contain
+                 NOT_SUPPORTED_MSG constant
+        """
+        is_jbod_capable = False
+
+        # Grab all the disks associated with the RAID controller
+        all_physical_disks = self.list_physical_disks()
+        physical_disks = [physical_disk for physical_disk in all_physical_disks
+                          if physical_disk.controller == raid_controller_fqdd]
+
+        # If there is a disk in the Non-RAID state, then the controller is JBOD
+        # capable
+        ready_disk = None
+        for physical_disk in physical_disks:
+            if physical_disk.raid_status == 'non-RAID':
+                is_jbod_capable = True
+                break
+            elif not ready_disk and physical_disk.raid_status == 'ready':
+                ready_disk = physical_disk
+
+        if not is_jbod_capable:
+            if not ready_disk:
+                msg = "Unable to find a disk in the Ready state"
+                raise exceptions.DRACRequestFailed(msg)
+
+            # Try moving a disk in the Ready state to JBOD mode
+            try:
+                self.convert_physical_disks(
+                    [ready_disk.id],
+                    False)
+                is_jbod_capable = True
+
+                # Flip the disk back to the Ready state.  This results in the
+                # pending value being reset to nothing, so it effectively
+                # undoes the last command and makes the check non-destructive
+                self.convert_physical_disks(
+                    [ready_disk.id],
+                    True)
+            except exceptions.DRACOperationFailed as ex:
+                # Fix for python 3, Exception.message no longer
+                # a valid attribute, str(ex) works for both 2.7
+                # and 3.x
+                if self.NOT_SUPPORTED_MSG in str(ex):
+                    pass
+                else:
+                    raise
+
+        return is_jbod_capable
