@@ -17,6 +17,7 @@ import random
 import requests_mock
 
 import dracclient.client
+from dracclient import constants
 from dracclient import exceptions
 import dracclient.resources.job
 from dracclient.resources import raid
@@ -33,6 +34,7 @@ class ClientRAIDManagementTestCase(base.BaseTest):
         super(ClientRAIDManagementTestCase, self).setUp()
         self.drac_client = dracclient.client.DRACClient(
             **test_utils.FAKE_ENDPOINT)
+        self.raid_controller_fqdd = "RAID.Integrated.1-1"
 
     @mock.patch.object(dracclient.client.WSManClient,
                        'wait_until_idrac_is_ready', spec_set=True,
@@ -45,7 +47,8 @@ class ClientRAIDManagementTestCase(base.BaseTest):
             manufacturer='DELL',
             model='PERC H710 Mini',
             primary_status='ok',
-            firmware_version='21.3.0-0009')
+            firmware_version='21.3.0-0009',
+            bus='1')
 
         mock_requests.post(
             'https://1.2.3.4:443/wsman',
@@ -102,7 +105,9 @@ class ClientRAIDManagementTestCase(base.BaseTest):
             serial_number='S0M3EY2Z',
             firmware_version='LS0A',
             status='ok',
-            raid_status='ready')
+            raid_status='ready',
+            sas_address='5000C5007764F409',
+            device_protocol=None)
 
         mock_requests.post(
             'https://1.2.3.4:443/wsman',
@@ -130,7 +135,38 @@ class ClientRAIDManagementTestCase(base.BaseTest):
             serial_number='S0M3EY3Z',
             firmware_version='LS0B',
             status='ok',
-            raid_status='ready')
+            raid_status='ready',
+            sas_address='5000C5007764F409',
+            device_protocol=None)
+
+        mock_requests.post(
+            'https://1.2.3.4:443/wsman',
+            text=test_utils.RAIDEnumerations[uris.DCIM_PhysicalDiskView]['ok'])
+
+        self.assertIn(expected_physical_disk,
+                      self.drac_client.list_physical_disks())
+
+    @mock.patch.object(dracclient.client.WSManClient,
+                       'wait_until_idrac_is_ready', spec_set=True,
+                       autospec=True)
+    def test_list_physical_disks_nvme(self, mock_requests,
+                                      mock_wait_until_idrac_is_ready):
+        expected_physical_disk = raid.PhysicalDisk(
+            id='Disk.Bay.20:Enclosure.Internal.0-1:PCIeExtender.Slot.1',
+            description='PCIe SSD in Slot 20 in Bay 1',
+            controller='PCIeExtender.Slot.1',
+            manufacturer='SAMSUNG',
+            model='Dell Express Flash PM1725a 800GB SFF',
+            media_type='ssd',
+            interface_type='pcie',
+            size_mb=763097,
+            free_size_mb=None,
+            serial_number='S39YNX0JB02343',
+            firmware_version='1.0.4',
+            status='unknown',
+            raid_status=None,
+            sas_address=None,
+            device_protocol='NVMe-MI1.0')
 
         mock_requests.post(
             'https://1.2.3.4:443/wsman',
@@ -162,13 +198,19 @@ class ClientRAIDManagementTestCase(base.BaseTest):
                               'SystemName': 'DCIM:ComputerSystem',
                               'Name': 'DCIM:RAIDService'}
         expected_properties = {'PDArray': [device_fqdd]}
+        mock_invoke.return_value = lxml.etree.fromstring(
+            test_utils.RAIDInvocations[uris.DCIM_RAIDService][
+                'ConvertToRAID']['ok'])
 
         result = self.drac_client.convert_physical_disks(
             raid_controller='controller',
             physical_disks=[device_fqdd],
             raid_enable=True)
 
-        self.assertEqual({'commit_required': False}, result)
+        self.assertEqual({'commit_required': True,
+                          'is_commit_required': True,
+                          'is_reboot_required': constants.RebootRequired.true},
+                         result)
         mock_invoke.assert_called_once_with(
             mock.ANY, uris.DCIM_RAIDService, expected_invocation,
             expected_selectors, expected_properties,
@@ -188,13 +230,19 @@ class ClientRAIDManagementTestCase(base.BaseTest):
                               'SystemName': 'DCIM:ComputerSystem',
                               'Name': 'DCIM:RAIDService'}
         expected_properties = {'PDArray': device_list}
+        mock_invoke.return_value = lxml.etree.fromstring(
+            test_utils.RAIDInvocations[uris.DCIM_RAIDService][
+                'ConvertToRAID']['ok'])
 
         result = self.drac_client.convert_physical_disks(
             raid_controller='controller',
             physical_disks=device_list,
             raid_enable=True)
 
-        self.assertEqual({'commit_required': False}, result)
+        self.assertEqual({'commit_required': True,
+                          'is_commit_required': True,
+                          'is_reboot_required': constants.RebootRequired.true},
+                         result)
         mock_invoke.assert_called_once_with(
             mock.ANY, uris.DCIM_RAIDService, expected_invocation,
             expected_selectors, expected_properties,
@@ -212,13 +260,19 @@ class ClientRAIDManagementTestCase(base.BaseTest):
                               'SystemName': 'DCIM:ComputerSystem',
                               'Name': 'DCIM:RAIDService'}
         expected_properties = {'PDArray': [device_fqdd]}
+        mock_invoke.return_value = lxml.etree.fromstring(
+            test_utils.RAIDInvocations[uris.DCIM_RAIDService][
+                'ConvertToRAID']['ok'])
 
         result = self.drac_client.convert_physical_disks(
             raid_controller='controller',
             physical_disks=[device_fqdd],
             raid_enable=False)
 
-        self.assertEqual({'commit_required': False}, result)
+        self.assertEqual({'commit_required': True,
+                          'is_commit_required': True,
+                          'is_reboot_required': constants.RebootRequired.true},
+                         result)
         mock_invoke.assert_called_once_with(
             mock.ANY, uris.DCIM_RAIDService, expected_invocation,
             expected_selectors, expected_properties,
@@ -239,43 +293,19 @@ class ClientRAIDManagementTestCase(base.BaseTest):
                               'SystemName': 'DCIM:ComputerSystem',
                               'Name': 'DCIM:RAIDService'}
         expected_properties = {'PDArray': device_list}
+        mock_invoke.return_value = lxml.etree.fromstring(
+            test_utils.RAIDInvocations[uris.DCIM_RAIDService][
+                'ConvertToRAID']['ok'])
 
         result = self.drac_client.convert_physical_disks(
             raid_controller='controller',
             physical_disks=device_list,
             raid_enable=False)
 
-        self.assertEqual({'commit_required': False}, result)
-        mock_invoke.assert_called_once_with(
-            mock.ANY, uris.DCIM_RAIDService, expected_invocation,
-            expected_selectors, expected_properties,
-            expected_return_value=utils.RET_SUCCESS)
-
-    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
-                       spec_set=True, autospec=True)
-    def test_convert_physical_disks_ok(self, mock_requests, mock_invoke):
-        '''Convert a number of disks to RAID mode and check the return value'''
-        device_list = []
-        for i in range(0, random.randint(2, 10)):
-            device_list += self._random_fqdd()
-
-        expected_invocation = 'ConvertToRAID'
-        expected_selectors = {'SystemCreationClassName': 'DCIM_ComputerSystem',
-                              'CreationClassName': 'DCIM_RAIDService',
-                              'SystemName': 'DCIM:ComputerSystem',
-                              'Name': 'DCIM:RAIDService'}
-        expected_properties = {'PDArray': device_list}
-
-        mock_invoke.return_value = lxml.etree.fromstring(
-            test_utils.RAIDInvocations[uris.DCIM_RAIDService][
-                expected_invocation]['ok'])
-
-        result = self.drac_client.convert_physical_disks(
-            raid_controller='controller',
-            physical_disks=device_list,
-            raid_enable=True)
-
-        self.assertEqual({'commit_required': True}, result)
+        self.assertEqual({'commit_required': True,
+                          'is_commit_required': True,
+                          'is_reboot_required': constants.RebootRequired.true},
+                         result)
         mock_invoke.assert_called_once_with(
             mock.ANY, uris.DCIM_RAIDService, expected_invocation,
             expected_selectors, expected_properties,
@@ -318,7 +348,10 @@ class ClientRAIDManagementTestCase(base.BaseTest):
             raid_controller='controller', physical_disks=['disk1', 'disk2'],
             raid_level='1', size_mb=42)
 
-        self.assertEqual({'commit_required': True}, result)
+        self.assertEqual({'commit_required': True,
+                          'is_commit_required': True,
+                          'is_reboot_required': constants.RebootRequired.true},
+                         result)
         mock_invoke.assert_called_once_with(
             mock.ANY, uris.DCIM_RAIDService, 'CreateVirtualDisk',
             expected_selectors, expected_properties,
@@ -348,7 +381,10 @@ class ClientRAIDManagementTestCase(base.BaseTest):
             raid_level='1', size_mb=42, disk_name='name', span_length=2,
             span_depth=3)
 
-        self.assertEqual({'commit_required': True}, result)
+        self.assertEqual({'commit_required': True,
+                          'is_commit_required': True,
+                          'is_reboot_required': constants.RebootRequired.true},
+                         result)
         mock_invoke.assert_called_once_with(
             mock.ANY, uris.DCIM_RAIDService, 'CreateVirtualDisk',
             expected_selectors, expected_properties,
@@ -405,6 +441,24 @@ class ClientRAIDManagementTestCase(base.BaseTest):
             physical_disks=['disk1', 'disk2'], raid_level='1', size_mb=None,
             disk_name='name', span_length=2, span_depth=3)
 
+    # This test is specifically for support of creating a RAID1 on a Dell BOSS
+    # card.  It requires that size_mb is set to 0
+    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
+                       spec_set=True, autospec=True)
+    def test_create_virtual_disk_0_size(self, mock_requests, mock_invoke):
+        mock_invoke.return_value = lxml.etree.fromstring(
+            test_utils.RAIDInvocations[uris.DCIM_RAIDService][
+                'CreateVirtualDisk']['ok'])
+
+        self.drac_client.create_virtual_disk(
+            raid_controller='controller',
+            physical_disks=['disk1', 'disk2'],
+            raid_level='1',
+            size_mb=0,
+            disk_name='name',
+            span_length=1,
+            span_depth=2)
+
     def test_create_virtual_disk_invalid_size(self, mock_requests):
         self.assertRaises(
             exceptions.InvalidParameterValue,
@@ -440,7 +494,10 @@ class ClientRAIDManagementTestCase(base.BaseTest):
 
         result = self.drac_client.delete_virtual_disk('disk1')
 
-        self.assertEqual({'commit_required': True}, result)
+        self.assertEqual({'commit_required': True,
+                          'is_commit_required': True,
+                          'is_reboot_required': constants.RebootRequired.true},
+                         result)
         mock_invoke.assert_called_once_with(
             mock.ANY, uris.DCIM_RAIDService, 'DeleteVirtualDisk',
             expected_selectors, expected_properties,
@@ -469,7 +526,8 @@ class ClientRAIDManagementTestCase(base.BaseTest):
         mock_create_config_job.assert_called_once_with(
             mock.ANY, resource_uri=uris.DCIM_RAIDService,
             cim_creation_class_name='DCIM_RAIDService',
-            cim_name='DCIM:RAIDService', target='controller', reboot=False)
+            cim_name='DCIM:RAIDService', target='controller', reboot=False,
+            start_time='TIME_NOW')
 
     @mock.patch.object(dracclient.resources.job.JobManagement,
                        'create_config_job', spec_set=True, autospec=True)
@@ -480,7 +538,39 @@ class ClientRAIDManagementTestCase(base.BaseTest):
         mock_create_config_job.assert_called_once_with(
             mock.ANY, resource_uri=uris.DCIM_RAIDService,
             cim_creation_class_name='DCIM_RAIDService',
-            cim_name='DCIM:RAIDService', target='controller', reboot=True)
+            cim_name='DCIM:RAIDService', target='controller', reboot=True,
+            start_time='TIME_NOW')
+
+    @mock.patch.object(dracclient.resources.job.JobManagement,
+                       'create_config_job', spec_set=True, autospec=True)
+    def test_commit_pending_raid_changes_with_start_time(
+            self, mock_requests,
+            mock_create_config_job):
+        timestamp = '20140924140201'
+        self.drac_client.commit_pending_raid_changes('controller',
+                                                     start_time=timestamp)
+
+        mock_create_config_job.assert_called_once_with(
+            mock.ANY, resource_uri=uris.DCIM_RAIDService,
+            cim_creation_class_name='DCIM_RAIDService',
+            cim_name='DCIM:RAIDService', target='controller', reboot=False,
+            start_time=timestamp)
+
+    @mock.patch.object(dracclient.resources.job.JobManagement,
+                       'create_config_job', spec_set=True, autospec=True)
+    def test_commit_pending_raid_changes_with_reboot_and_start_time(
+            self, mock_requests,
+            mock_create_config_job):
+        timestamp = '20140924140201'
+        self.drac_client.commit_pending_raid_changes('controller',
+                                                     reboot=True,
+                                                     start_time=timestamp)
+
+        mock_create_config_job.assert_called_once_with(
+            mock.ANY, resource_uri=uris.DCIM_RAIDService,
+            cim_creation_class_name='DCIM_RAIDService',
+            cim_name='DCIM:RAIDService', target='controller', reboot=True,
+            start_time=timestamp)
 
     @mock.patch.object(dracclient.resources.job.JobManagement,
                        'delete_pending_config', spec_set=True, autospec=True)
@@ -492,3 +582,117 @@ class ClientRAIDManagementTestCase(base.BaseTest):
             mock.ANY, resource_uri=uris.DCIM_RAIDService,
             cim_creation_class_name='DCIM_RAIDService',
             cim_name='DCIM:RAIDService', target='controller')
+
+    @mock.patch.object(dracclient.client.WSManClient,
+                       'wait_until_idrac_is_ready', spec_set=True,
+                       autospec=True)
+    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
+                       'convert_physical_disks',
+                       return_value={}, spec_set=True,
+                       autospec=True)
+    def test_raid_controller_jbod_capable(self, mock_requests,
+                                          mock_wait_until_idrac_is_ready,
+                                          mock_convert_physical_disks):
+
+        mock_requests.post(
+            'https://1.2.3.4:443/wsman',
+            text=test_utils.RAIDEnumerations[uris.DCIM_PhysicalDiskView]['ok'])
+
+        is_jbod = self.drac_client.is_jbod_capable(self.raid_controller_fqdd)
+
+        self.assertTrue(is_jbod, msg="is_jbod is true")
+
+    @mock.patch.object(dracclient.client.WSManClient,
+                       'wait_until_idrac_is_ready', spec_set=True,
+                       autospec=True)
+    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
+                       'convert_physical_disks',
+                       return_value={}, spec_set=True,
+                       autospec=True)
+    def test_raid_controller_jbod_non_raid(self, mock_requests,
+                                           mock_wait_until_idrac_is_ready,
+                                           mock_convert_physical_disks):
+
+        pdv = test_utils.RAIDEnumerations[uris.DCIM_PhysicalDiskView]['ok']
+        # change to non-RAID value
+        pdv = pdv.replace("<n1:RaidStatus>1</n1:RaidStatus>",
+                          "<n1:RaidStatus>8</n1:RaidStatus>")
+
+        mock_requests.post(
+            'https://1.2.3.4:443/wsman',
+            text=pdv)
+
+        is_jbod = self.drac_client.is_jbod_capable(self.raid_controller_fqdd)
+
+        self.assertTrue(is_jbod, msg="is_jbod is true")
+
+    @mock.patch.object(dracclient.client.WSManClient,
+                       'wait_until_idrac_is_ready', spec_set=True,
+                       autospec=True)
+    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
+                       'convert_physical_disks',
+                       return_value={}, spec_set=True,
+                       autospec=True)
+    def test_raid_controller_jbod_unknown(self, mock_requests,
+                                          mock_wait_until_idrac_is_ready,
+                                          mock_convert_physical_disks):
+
+        is_jbod = False
+        pdv = test_utils.RAIDEnumerations[uris.DCIM_PhysicalDiskView]['ok']
+        # change to non-RAID value
+        pdv = pdv.replace("<n1:RaidStatus>1</n1:RaidStatus>",
+                          "<n1:RaidStatus>0</n1:RaidStatus>")
+
+        mock_requests.post(
+            'https://1.2.3.4:443/wsman',
+            text=pdv)
+        self.assertRaises(exceptions.DRACRequestFailed,
+                          self.drac_client.is_jbod_capable,
+                          self.raid_controller_fqdd)
+        self.assertFalse(is_jbod, msg="is_jbod is false")
+
+    @mock.patch.object(dracclient.client.WSManClient,
+                       'wait_until_idrac_is_ready', spec_set=True,
+                       autospec=True)
+    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
+                       'convert_physical_disks',
+                       spec_set=True,
+                       autospec=True)
+    def test_raid_controller_jbod_not_supported(self,
+                                                mock_requests,
+                                                mock_convert_physical_disks,
+                                                mock_wait_idrac_is_ready):
+
+        msg = " operation is not supported on th"
+        exc = exceptions.DRACOperationFailed(drac_messages=msg)
+        mock_convert_physical_disks.side_effect = exc
+
+        mock_requests.post(
+            'https://1.2.3.4:443/wsman',
+            text=test_utils.RAIDEnumerations[uris.DCIM_PhysicalDiskView]['ok'])
+
+        is_jbod = self.drac_client.is_jbod_capable(self.raid_controller_fqdd)
+        self.assertFalse(is_jbod, msg="is_jbod is false")
+
+    @mock.patch.object(dracclient.client.WSManClient,
+                       'wait_until_idrac_is_ready', spec_set=True,
+                       autospec=True)
+    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
+                       'convert_physical_disks',
+                       spec_set=True,
+                       autospec=True)
+    def test_raid_controller_jbod_ex_no_match(self,
+                                              mock_requests,
+                                              mock_convert_physical_disks,
+                                              mock_wait_until_idrac_is_ready):
+
+        mock_requests.post(
+            'https://1.2.3.4:443/wsman',
+            text=test_utils.RAIDEnumerations[uris.DCIM_PhysicalDiskView]['ok'])
+        msg = "NON_MATCHING_MESSAGE"
+        exc = exceptions.DRACOperationFailed(drac_messages=msg)
+        mock_convert_physical_disks.side_effect = exc
+
+        self.assertRaises(
+            exceptions.DRACOperationFailed,
+            self.drac_client.is_jbod_capable, self.raid_controller_fqdd)
